# Authors: Ahmet Bayraktar
#          Stan Leisink
#          Frits Hermans

import abc
from typing import List, Tuple, Optional, Union

import numpy as np
from pyspark.ml.feature import CountVectorizer, VectorAssembler, MinHashLSH
from pyspark.ml.linalg import DenseVector, SparseVector
from pyspark.sql import DataFrame, functions as F, types as T

from spark_matcher.config import MINHASHING_MAXDF, MINHASHING_VOCABSIZE
from spark_matcher.table_checkpointer import TableCheckpointer


class Sampler:
    """
    Sampler base class to generate pairs for training.

    Args:
        col_names: list of column names to use for matching
        n_train_samples: nr of pair samples to be created for training
        table_checkpointer: table_checkpointer object to store cached tables
    """
    def __init__(self, col_names: List[str], n_train_samples: int, table_checkpointer: TableCheckpointer) -> DataFrame:
        self.col_names = col_names
        self.n_train_samples = n_train_samples
        self.table_checkpointer = table_checkpointer

    @abc.abstractmethod
    def create_pairs_table(self, sdf_1: DataFrame, sdf_2: DataFrame) -> DataFrame:
        pass


class HashSampler(Sampler):
    def __init__(self, table_checkpointer: TableCheckpointer, col_names: List[str], n_train_samples: int,
                 threshold: float = 0.5, num_hash_tables: int = 10) -> DataFrame:
        """
        Sampler class to generate pairs using MinHashLSH method by selecting pairs that are more likely to be a match.

        Args:
            threshold: for the distance of hashed pairs, values below threshold will be returned
                       (threshold being equal to 1.0 will return all pairs with at least one common shingle)
            num_hash_tables: number of hashes to be applied
            table_checkpointer: table_checkpointer object to store cached tables
        Returns:
            A spark dataframe which contains all selected pairs for training
        """
        super().__init__(col_names, n_train_samples, table_checkpointer)
        self.threshold = threshold
        self.num_hash_tables = num_hash_tables

    @staticmethod
    @F.udf(T.BooleanType())
    def is_non_zero_vector(vector):
        """
        Check if a vector has at least 1 non-zero entry. This function can deal with dense or sparse vectors. This is
        needed as the VectorAssembler can return dense or sparse vectors, dependent on what is more memory efficient.

        Args:
            vector: vector

        Returns:
            boolean whether a vector has at least 1 non-zero entry

        """
        if isinstance(vector, DenseVector):
            return bool(len(vector))
        if isinstance(vector, SparseVector):
            return vector.indices.size > 0

    def _vectorize(self, col_names: List[str], sdf_1: DataFrame, sdf_2: Optional[DataFrame] = None,
                   max_df: Union[float, int] = MINHASHING_MAXDF, vocab_size: int = MINHASHING_VOCABSIZE) -> Union[
        Tuple[DataFrame, DataFrame], DataFrame]:
        """
        This function is used to vectorize word based features. First, given dataframes are united to cover all feature
        space. Given columns are converted into a vector. `sdf_2` is only required for record matching, for
        deduplication only `sdf_1` is required. To prevent too many (redundant) matches on frequently occurring tokens,
        the maximum document frequency for vectorization is limited by maxDF.

        Args:
            col_names: list of column names to create feature vectors
            sdf_1: spark dataframe
            sdf_2: Optional: spark dataframe
            max_df: max document frequency to use for count vectorization
            vocab_size: vocabulary size of vectorizer

        Returns:
            Adds a 'features' column to given spark dataframes

        """
        if sdf_2:
            # creating a source column to separate tables after vectorization is completed
            sdf_1 = sdf_1.withColumn('source', F.lit('sdf_1'))
            sdf_2 = sdf_2.withColumn('source', F.lit('sdf_2'))
            sdf_merged = sdf_1.unionByName(sdf_2)
        else:
            sdf_merged = sdf_1

        for col in col_names:
            # converting strings into an array
            sdf_merged = sdf_merged.withColumn(f'{col}_array', F.split(F.col(col), ' '))

            # setting CountVectorizer
            # minDF is set to 2 because if a token occurs only once there is no other occurrence to match it to
            # the maximum document frequency may not be smaller than the minimum document frequency which is assured
            # below
            n_rows = sdf_merged.count()
            if isinstance(max_df, float) and (n_rows * max_df < 2):
                max_df = 10
            cv = CountVectorizer(binary=True, minDF=2, maxDF=max_df, vocabSize=vocab_size, inputCol=f"{col}_array",
                                 outputCol=f"{col}_vector")

            # creating vectors
            model = cv.fit(sdf_merged)
            sdf_merged = model.transform(sdf_merged)
        sdf_merged = self.table_checkpointer(sdf_merged, "cached_vectorized_table")

        # in case of col_names contains multiple columns, merge all vectors into one
        vec_assembler = VectorAssembler(outputCol="features")
        vec_assembler.setInputCols([f'{col}_vector' for col in col_names])

        sdf_merged = vec_assembler.transform(sdf_merged)

        # breaking the lineage is found to be required below
        sdf_merged = self.table_checkpointer(
            sdf_merged.filter(HashSampler.is_non_zero_vector('features')), 'minhash_vectorized')

        if sdf_2:
            sdf_1_vectorized = sdf_merged.filter(F.col('source') == 'sdf_1').select(*col_names, 'features')
            sdf_2_vectorized = sdf_merged.filter(F.col('source') == 'sdf_2').select(*col_names, 'features')
            return sdf_1_vectorized, sdf_2_vectorized
        sdf_1_vectorized = sdf_merged.select(*col_names, 'features')
        return sdf_1_vectorized

    @staticmethod
    def _apply_min_hash(sdf_1: DataFrame, sdf_2: DataFrame, col_names: List[str],
                        threshold: float, num_hash_tables: int):
        """
        This function is used to apply MinHasLSH technique to calculate the Jaccard distance between feature vectors.

        Args:
            sdf_1: spark dataframe with the `features` column and the column `row_number_1`
            sdf_2: spark dataframe with the `features` column and the column `row_number_2`
            col_names: list of column names to create feature vectors

        Returns:
            Creates one spark dataframe that contains all pairs that has a smaller Jaccard distance than threshold
        """
        mh = MinHashLSH(inputCol="features", outputCol="hashes", seed=42, numHashTables=num_hash_tables)

        model = mh.fit(sdf_1)
        model.transform(sdf_1)

        sdf_distances = (
            model
                .approxSimilarityJoin(sdf_1, sdf_2, threshold, distCol="JaccardDistance")
                .select(*[F.col(f'datasetA.{col}').alias(f'{col}_1') for col in col_names],
                        *[F.col(f'datasetB.{col}').alias(f'{col}_2') for col in col_names],
                        F.col('JaccardDistance').alias('distance'),
                        F.col('datasetA.row_number_1').alias('row_number_1'),
                        F.col('datasetB.row_number_2').alias('row_number_2'))
        )
        return sdf_distances

    def create_pairs_table(self, sdf_1: DataFrame, sdf_2: Optional[DataFrame] = None) -> DataFrame:
        """
        Create hashed pairs that are used for training. `sdf_2` is only required for record matching, for deduplication
        only `sdf_1` is required.

        Args:
            sdf_1: Spark dataframe containing the first table to with the input should be matched
            sdf_2: Optional: Spark dataframe containing the second table that should be matched to the first table

        Returns:
            Spark dataframe that contains sampled pairs selected with MinHashLSH technique

        """
        if sdf_2:
            sdf_1_vectorized, sdf_2_vectorized = self._vectorize(self.col_names, sdf_1, sdf_2)
            sdf_1_vectorized = self.table_checkpointer(
                                sdf_1_vectorized.withColumn('row_number_1', F.monotonically_increasing_id()),
                                checkpoint_name='sdf_1_vectorized')
            sdf_2_vectorized = self.table_checkpointer(
                                sdf_2_vectorized.withColumn('row_number_2', F.monotonically_increasing_id()),
                                checkpoint_name='sdf_2_vectorized')
        else:
            sdf_1_vectorized = self._vectorize(self.col_names, sdf_1)
            sdf_1_vectorized = self.table_checkpointer(
                                sdf_1_vectorized.withColumn('row_number_1', F.monotonically_increasing_id()),
                                checkpoint_name='sdf_1_vectorized')
            sdf_2_vectorized = sdf_1_vectorized.alias('sdf_2_vectorized') # matched with itself for deduplication
            sdf_2_vectorized = sdf_2_vectorized.withColumnRenamed('row_number_1', 'row_number_2')

        sdf_distances = self._apply_min_hash(sdf_1_vectorized, sdf_2_vectorized, self.col_names,
                                             self.threshold, self.num_hash_tables)

        # for deduplication we remove identical pairs like (a,a) and duplicates of pairs like (a,b) and (b,a)
        if not sdf_2:
            sdf_distances = sdf_distances.filter(F.col('row_number_1') < F.col('row_number_2'))

        hashed_pairs_table = (
            sdf_distances
            .sort('distance')
            .limit(self.n_train_samples)
            .drop('distance', 'row_number_1', 'row_number_2')
            )

        return hashed_pairs_table


class RandomSampler(Sampler):
    def __init__(self, table_checkpointer: TableCheckpointer, col_names: List[str], n_train_samples: int) -> DataFrame:
        """
        Sampler class to generate randomly selected pairs

        Returns:
            A spark dataframe which contains randomly selected pairs for training
        """
        super().__init__(col_names, n_train_samples, table_checkpointer)

    def create_pairs_table(self, sdf_1: DataFrame, sdf_2: Optional[DataFrame] = None) -> DataFrame:
        """
        Create random pairs that are used for training. `sdf_2` is only required for record matching, for deduplication
        only `sdf_1` is required.

        Args:
            sdf_1: Spark dataframe containing the first table to with the input should be matched
            sdf_2: Optional: Spark dataframe containing the second table that should be matched to the first table

        Returns:
            Spark dataframe that contains randomly sampled pairs

        """
        if sdf_2:
            return self._create_pairs_table_matcher(sdf_1, sdf_2)
        return self._create_pairs_table_deduplicator(sdf_1)

    def _create_pairs_table_matcher(self, sdf_1: DataFrame, sdf_2: DataFrame) -> DataFrame:
        """
        Create random pairs that are used for training.

        If the first table and the second table are equally large, we take the square root of `n_train_samples` from
        both tables to create a pairs table containing `n_train_samples` rows. If one of the table is much smaller than
        the other, all rows of the smallest table are taken and the number of the sample from the larger table is chosen
        such that the total number of pairs is `n_train_samples`.

        Args:
            sdf_1: Spark dataframe containing the first table to with the input should be matched
            sdf_2: Spark dataframe containing the second table that should be matched to the first table

        Returns:
            Spark dataframe with randomly sampled pairs to compared during training
        """
        sdf_1_count, sdf_2_count = sdf_1.count(), sdf_2.count()

        sample_size_small_table = min([int(self.n_train_samples ** 0.5), min(sdf_1_count, sdf_2_count)])
        sample_size_large_table = self.n_train_samples // sample_size_small_table

        if sdf_1_count > sdf_2_count:
            fraction_sdf_1 = sample_size_large_table / sdf_1_count
            fraction_sdf_2 = sample_size_small_table / sdf_2_count
            # below the `fraction` is slightly increased (but capped at 1.) and a `limit()` is applied to assure that
            # the exact number of required samples is obtained
            sdf_1_sample = (sdf_1.sample(withReplacement=False, fraction=min([1., 1.5 * fraction_sdf_1]))
                            .limit(sample_size_large_table))
            sdf_2_sample = (sdf_2.sample(withReplacement=False, fraction=min([1., 1.5 * fraction_sdf_2]))
                            .limit(sample_size_small_table))
        else:
            fraction_sdf_1 = sample_size_small_table / sdf_1_count
            fraction_sdf_2 = sample_size_large_table / sdf_2_count
            sdf_1_sample = (sdf_1.sample(withReplacement=False, fraction=min([1., 1.5 * fraction_sdf_1]))
                            .limit(sample_size_small_table))
            sdf_2_sample = (sdf_2.sample(withReplacement=False, fraction=min([1., 1.5 * fraction_sdf_2]))
                            .limit(sample_size_large_table))

        for col in self.col_names:
            sdf_1_sample = sdf_1_sample.withColumnRenamed(col, col + "_1")
            sdf_2_sample = sdf_2_sample.withColumnRenamed(col, col + "_2")

        random_pairs_table = sdf_1_sample.crossJoin(sdf_2_sample)

        return random_pairs_table

    def _create_pairs_table_deduplicator(self, sdf: DataFrame) -> DataFrame:
        """
        Create randomly selected pairs for deduplication.

        Args:
            sdf: Spark dataframe containing rows from which pairs need to be created

        Returns:
            Spark dataframe with the randomly selected pairs

        """
        n_samples_required = int(1.5 * np.ceil((self.n_train_samples * 2) ** 0.5))
        fraction_samples_required = min([n_samples_required / sdf.count(), 1.])
        sample = sdf.sample(withReplacement=False, fraction=fraction_samples_required)
        sample = self.table_checkpointer(sample.withColumn('row_id', F.monotonically_increasing_id()),
                                                          checkpoint_name='random_pairs_deduplicator')
        sample_1, sample_2 = sample, sample
        for col in self.col_names + ['row_id']:
            sample_1 = sample_1.withColumnRenamed(col, col + "_1")
            sample_2 = sample_2.withColumnRenamed(col, col + "_2")

        pairs_table = (sample_1
                       .crossJoin(sample_2)
                       .filter(F.col('row_id_1') < F.col('row_id_2'))
                       .limit(self.n_train_samples)
                       .drop('row_id_1', 'row_id_2'))

        return pairs_table



# Authors: Ahmet Bayraktar
#          Stan Leisink
#          Frits Hermans

from typing import Tuple, Optional, Union
from pkg_resources import resource_filename

import pandas as pd

from pyspark.sql import SparkSession, DataFrame


def load_data(spark: SparkSession, kind: Optional[str] = 'voters') -> Union[Tuple[DataFrame, DataFrame], DataFrame]:
    """
    Load examples datasets to be used to experiment with `spark-matcher`. For matching problems, set `kind` to `voters`
    for North Carolina voter registry data or `library` for bibliography data. For deduplication problems, set `kind`
    to `stoxx50` for EuroStoxx 50 company names and addresses.

    Voter data:
        - provided by Prof. Erhard Rahm
          https://dbs.uni-leipzig.de/research/projects/object_matching/benchmark_datasets_for_entity_resolution

    Library data:
        - DBLP bibliography, http://www.informatik.uni-trier.de/~ley/db/index.html
        - ACM Digital Library, http://portal.acm.org/portal.cfm

    Args:
        spark: Spark session
        kind: kind of data: `voters`, `library` or `stoxx50`

    Returns:
        two Spark dataframes for `voters` or `library`, a single dataframe for `stoxx50`

    """
    if kind == 'library':
        return _load_data_library(spark)
    if kind == 'voters':
        return _load_data_voters(spark)
    if kind == 'stoxx50':
        return _load_data_stoxx50(spark)
    else:
        raise ValueError('`kind` must be `library`, `voters` or `stoxx50`')


def _load_data_library(spark: SparkSession) -> Tuple[DataFrame, DataFrame]:
    """
    Load examples datasets to be used to experiment with `spark-matcher`. Two Spark dataframe are returned with the
    same columns:

    - DBLP bibliography, http://www.informatik.uni-trier.de/~ley/db/index.html
    - ACM Digital Library, http://portal.acm.org/portal.cfm

    Args:
        spark: Spark session

    Returns:
        Spark dataframe for DBLP data and a Spark dataframe for ACM data

    """
    file_path_acm = resource_filename('spark_matcher.data', 'acm.csv')
    file_path_dblp = resource_filename('spark_matcher.data', 'dblp.csv')
    acm_pdf = pd.read_csv(file_path_acm)
    dblp_pdf = pd.read_csv(file_path_dblp, encoding="ISO-8859-1")

    for col in acm_pdf.select_dtypes('object').columns:
        acm_pdf[col] = acm_pdf[col].fillna("")
    for col in dblp_pdf.select_dtypes('object').columns:
        dblp_pdf[col] = dblp_pdf[col].fillna("")

    acm_sdf = spark.createDataFrame(acm_pdf)
    dblp_sdf = spark.createDataFrame(dblp_pdf)
    return acm_sdf, dblp_sdf


def _load_data_voters(spark: SparkSession) -> Tuple[DataFrame, DataFrame]:
    """
    Voters data is based on the North Carolina voter registry and this dataset is provided by Prof. Erhard Rahm
    ('Comparative Evaluation of Distributed Clustering Schemes for Multi-source Entity Resolution'). Two Spark
    dataframe are returned with the same columns.

    Args:
        spark: Spark session

    Returns:
        two Spark dataframes containing voter data

    """
    file_path_voters_1 = resource_filename('spark_matcher.data', 'voters_1.csv')
    file_path_voters_2 = resource_filename('spark_matcher.data', 'voters_2.csv')
    voters_1_pdf = pd.read_csv(file_path_voters_1)
    voters_2_pdf = pd.read_csv(file_path_voters_2)

    voters_1_sdf = spark.createDataFrame(voters_1_pdf)
    voters_2_sdf = spark.createDataFrame(voters_2_pdf)
    return voters_1_sdf, voters_2_sdf


def _load_data_stoxx50(spark: SparkSession) -> DataFrame:
    """
    The Stoxx50 dataset contains a single column containing the concatenation of Eurostoxx 50 company names and
    addresses. This dataset is created by the developers of spark_matcher.

    Args:
        spark: Spark session

    Returns:
        Spark dataframe containing Eurostoxx 50 names and addresses

    """
    file_path_stoxx50 = resource_filename('spark_matcher.data', 'stoxx50.csv')
    stoxx50_pdf = pd.read_csv(file_path_stoxx50)

    stoxx50_sdf = spark.createDataFrame(stoxx50_pdf)
    return stoxx50_sdf



# Authors: Ahmet Bayraktar
#          Stan Leisink
#          Frits Hermans

from typing import Optional, List, Dict

from pyspark.sql import DataFrame, SparkSession, functions as F, types as T
from sklearn.exceptions import NotFittedError

from spark_matcher.blocker.blocking_rules import BlockingRule
from spark_matcher.deduplicator.connected_components_calculator import ConnectedComponentsCalculator
from spark_matcher.deduplicator.hierarchical_clustering import apply_deduplication
from spark_matcher.matching_base.matching_base import MatchingBase
from spark_matcher.scorer.scorer import Scorer
from spark_matcher.table_checkpointer import TableCheckpointer


class Deduplicator(MatchingBase):
    """
    Deduplicator class to apply deduplication. Provide either the column names `col_names` using the default string
    similarity metrics or explicitly define the string similarity metrics in a dict `field_info` as in the example
    below. If `blocking_rules` is left empty, default blocking rules are used. Otherwise, provide blocking rules as
    a list containing `BlockingRule` instances (see example below). The number of perfect matches used during
    training is set by `n_perfect_train_matches`.

    E.g.:

    from spark_matcher.blocker.blocking_rules import FirstNChars

    myDeduplicator = Deduplicator(spark_session, field_info={'name':[metric_function_1, metric_function_2],
                                                            'address:[metric_function_1, metric_function_3]},
                                                            blocking_rules=[FirstNChars('name', 3)])

    Args:
        spark_session: Spark session
        col_names: list of column names to use for matching
        field_info: dict of column names as keys and lists of string similarity metrics as values
        blocking_rules: list of `BlockingRule` instances
        table_checkpointer: pointer object to store cached tables
        checkpoint_dir: checkpoint directory if provided
        n_train_samples: nr of pair samples to be created for training
        ratio_hashed_samples: ratio of hashed samples to be created for training, rest is sampled randomly
        n_perfect_train_matches: nr of perfect matches used for training
        scorer: a Scorer object used for scoring pairs
        verbose: sets verbosity
        max_edges_clustering: max number of edges per component that enters clustering
        edge_filter_thresholds: list of score thresholds to use for filtering when components are too large
        cluster_score_threshold: threshold value between [0.0, 1.0], only pairs are put together in clusters if
                                 cluster similarity scores are >= cluster_score_threshold
    """
    def __init__(self, spark_session: SparkSession, col_names: Optional[List[str]] = None,
                 field_info: Optional[Dict] = None, blocking_rules: Optional[List[BlockingRule]] = None,
                 blocking_recall: float = 1.0, table_checkpointer: Optional[TableCheckpointer] = None,
                 checkpoint_dir: Optional[str] = None, n_perfect_train_matches=1, n_train_samples: int = 100_000,
                 ratio_hashed_samples: float = 0.5, scorer: Optional[Scorer] = None, verbose: int = 0,
                 max_edges_clustering: int = 500_000,
                 edge_filter_thresholds: List[float] = [0.45, 0.55, 0.65, 0.75, 0.85, 0.95],
                 cluster_score_threshold: float = 0.5):

        super().__init__(spark_session, table_checkpointer, checkpoint_dir, col_names, field_info, blocking_rules,
                         blocking_recall, n_perfect_train_matches, n_train_samples, ratio_hashed_samples, scorer,
                         verbose)

        self.fitted_ = False
        self.max_edges_clustering = max_edges_clustering
        self.edge_filter_thresholds = edge_filter_thresholds
        self.cluster_score_threshold = cluster_score_threshold
        # set the checkpoints directory for graphframes
        self.spark_session.sparkContext.setCheckpointDir('checkpoints/')

    def _create_predict_pairs_table(self, sdf_blocked: DataFrame) -> DataFrame:
        """
        This method performs an alias self-join on `sdf` to create the pairs table for prediction. `sdf` is joined to
        itself based on the `block_key` column. The result of this join is a pairs table.
        """
        sdf_blocked_1 = self._add_suffix_to_col_names(sdf_blocked, 1)
        sdf_blocked_1 = sdf_blocked_1.withColumnRenamed('row_number', 'row_number_1')
        sdf_blocked_2 = self._add_suffix_to_col_names(sdf_blocked, 2)
        sdf_blocked_2 = sdf_blocked_2.withColumnRenamed('row_number', 'row_number_2')

        pairs = (
            sdf_blocked_1
                .join(sdf_blocked_2, on='block_key', how='inner')
                .filter(F.col("row_number_1") < F.col("row_number_2"))
                .drop_duplicates(subset=[col + "_1" for col in self.col_names] + [col + "_2" for col in self.col_names])
        )
        return pairs

    def _map_distributed_identifiers_to_long(self, clustered_results: DataFrame) -> DataFrame:
        """
        Method to add a unique `entity_identifier` to the results from clustering

        Args:
            clustered_results: results from clustering
        Returns:
            Spark dataframe containing `row_number` and `entity_identifier`
        """
        long_entity_ids = (
            clustered_results
                .select('entity_identifier')
                .drop_duplicates()
                .withColumn('long_entity_identifier', F.monotonically_increasing_id())
        )
        long_entity_ids = self.table_checkpointer(long_entity_ids, 'cached_long_ids_table')

        clustered_results = (
            clustered_results
            .join(long_entity_ids, on='entity_identifier', how='left')
            .drop('entity_identifier')
            .withColumnRenamed('long_entity_identifier', 'entity_identifier')
        )
        return clustered_results

    def _distributed_deduplication(self, scored_pairs_with_components: DataFrame) -> DataFrame:
        schema = T.StructType([T.StructField('row_number', T.LongType(), True),
                               T.StructField('entity_identifier', T.StringType(), True)])

        clustered_results = (
            scored_pairs_with_components
            .select('row_number_1', 'row_number_2', 'score', 'component_id')
            .groupby('component_id')
            .applyInPandas(apply_deduplication(self.cluster_score_threshold) , schema=schema)
        )
        return self.table_checkpointer(clustered_results, "cached_clustered_results_table")

    @staticmethod
    def _add_singletons_entity_identifiers(result_sdf: DataFrame) -> DataFrame:
        """
        Function to add entity_identifier to entities (singletons) that are not combined with other entities into a
        deduplicated entity. If there are no singletons, the input table will be returned as it is.

        Args:
            result_sdf: Spark dataframe containing the result of deduplication where entities that are not deduplicated
            have a missing value in the `entity_identifier` column.

        Returns:
            Spark dataframe with `entity_identifier` values for all entities

        """
        if result_sdf.filter(F.col('entity_identifier').isNull()).count() == 0:
            return result_sdf
        start_cluster_id = (
            result_sdf
                .filter(F.col('entity_identifier').isNotNull())
                .select(F.max('entity_identifier'))
                .first()[0])

        singletons_entity_identifiers = (
            result_sdf
                .filter(F.col('entity_identifier').isNull())
                .select('row_number')
                .rdd
                .zipWithIndex()
                .toDF()
                .withColumn('row_number', F.col('_1').getItem("row_number"))
                .drop("_1")
                .withColumn('entity_identifier_singletons', F.col('_2') + start_cluster_id + 1)
                .drop("_2"))

        result_sdf = (
            result_sdf
                .join(singletons_entity_identifiers, on='row_number', how='left')
                .withColumn('entity_identifier',
                            F.when(F.col('entity_identifier').isNull(),
                                F.col('entity_identifier_singletons'))
                            .otherwise(F.col('entity_identifier')))
                .drop('entity_identifier_singletons'))

        return result_sdf

    def _create_deduplication_results(self, clustered_results: DataFrame, entities: DataFrame) -> DataFrame:
        """
        Joins deduplication results back to entity_table and adds identifiers to rows that are not deduplicated with
        other rows
        """
        deduplication_results = (
            entities
            .join(clustered_results, on='row_number', how='left')
        )

        deduplication_results = (
                self._add_singletons_entity_identifiers(deduplication_results)
                .drop('row_number', 'block_key')
        )
        return deduplication_results

    def _get_large_clusters(self, pairs_with_components: DataFrame) -> DataFrame:
        """
        Components that are too large after iteratively removing edges with a similarity score lower than the
        thresholds in `edge_filter_thresholds`, are considered to be one entity. For these the `component_id` is
        temporarily used as an `entity_identifier`.
        """
        large_component = (pairs_with_components.filter(F.col('is_in_big_component'))
                           .withColumn('entity_identifier', F.col('component_id')))

        large_cluster = (large_component.select('row_number_1', 'entity_identifier')
                         .withColumnRenamed('row_number_1', 'row_number')
                         .unionByName(large_component.select('row_number_2', 'entity_identifier')
                                      .withColumnRenamed('row_number_2', 'row_number'))
                         .drop_duplicates()
                         .persist())
        return large_cluster

    def predict(self, sdf: DataFrame, threshold: float = 0.5):
        """
        Method to predict on data used for training or new data.

        Args:
            sdf: table to be applied entity deduplication
            threshold: probability threshold for similarity score

        Returns:
            Spark dataframe with the deduplication result

        """
        if not self.fitted_:
            raise NotFittedError('The Deduplicator instance is not fitted yet. Call `fit` and train the instance.')

        sdf = self._simplify_dataframe_for_matching(sdf)
        sdf = sdf.withColumn('row_number', F.monotonically_increasing_id())
        # make sure the `row_number` is a fixed value. See the docs of `monotonically_increasing_id`, the function
        # depends on the partitioning of the table. Not fixing/storing the result will cause problems when the table is
        # recalculated under the hood during other operations. Saving it to disk and reading it in, breaks lineage and
        # forces the result to be deterministic in subsequent operations.
        sdf = self.table_checkpointer(sdf, "cached_entities_numbered_table")

        sdf_blocked = self.table_checkpointer(self.blocker.transform(sdf), "cached_block_table")
        pairs_table = self.table_checkpointer(self._create_predict_pairs_table(sdf_blocked), "cached_pairs_table")
        metrics_table = self.table_checkpointer(self._calculate_metrics(pairs_table), "cached_metrics_table")

        scores_table = (
            metrics_table
            .withColumn('score', self.scoring_learner.predict_proba(metrics_table['similarity_metrics']))
        )
        scores_table = self.table_checkpointer(scores_table, "cached_scores_table")

        scores_table_filtered = (
            scores_table
            .filter(F.col('score') >= threshold)
            .drop('block_key', 'similarity_metrics')
            )

        # create the subgraphs by using the connected components algorithm
        connected_components_calculator = ConnectedComponentsCalculator(scores_table_filtered,
                                                                        self.max_edges_clustering,
                                                                        self.edge_filter_thresholds,
                                                                        self.table_checkpointer)

        # calculate the component id for each scored-pairs
        scored_pairs_with_component_ids = connected_components_calculator._create_component_ids()

        large_cluster = self._get_large_clusters(scored_pairs_with_component_ids)

        # apply the distributed deduplication to components that are sufficiently small
        clustered_results = self._distributed_deduplication(
            scored_pairs_with_component_ids.filter(~F.col('is_in_big_component')))

        all_results = large_cluster.unionByName(clustered_results)

        # assign a unique entity_identifier to all components
        all_results_long = self._map_distributed_identifiers_to_long(all_results)
        all_results_long = self.table_checkpointer(all_results_long, "cached_clustered_results_with_long_ids_table")

        deduplication_results = self._create_deduplication_results(all_results_long, sdf)

        return deduplication_results



# Authors: Ahmet Bayraktar
#          Stan Leisink
#          Frits Hermans

from typing import List, Tuple, Optional, Union

from pyspark.sql import DataFrame, functions as F

from spark_matcher.blocker.blocking_rules import BlockingRule
from spark_matcher.table_checkpointer import TableCheckpointer


class BlockLearner:
    """
    Class to learn blocking rules from training data.

    Attributes:
        blocking_rules: list of `BlockingRule` objects that are taken into account during block learning
        recall: the minimum required percentage of training pairs that are covered by the learned blocking rules
        verbose: set verbosity
    """
    def __init__(self, blocking_rules: List[BlockingRule], recall: float,
                 table_checkpointer: Optional[TableCheckpointer] = None, verbose=0):
        self.table_checkpointer = table_checkpointer
        self.blocking_rules = blocking_rules
        self.recall = recall
        self.cover_blocking_rules = None
        self.fitted = False
        self.full_set = None
        self.full_set_size = None
        self.verbose = verbose

    def _greedy_set_coverage(self):
        """
        This method solves the `set cover problem` with a greedy algorithm. It identifies a subset of blocking-rules
        that cover `recall` * |full_set| (|full_set| stands for cardinality of the full_set)percent of all the elements
        in the original (full) set.
        """
        # sort the blocking rules to start the algorithm with the ones that have most coverage:
        _sorted_blocking_rules = sorted(self.blocking_rules, key=lambda bl: bl.training_coverage_size, reverse=True)

        self.cover_blocking_rules = [_sorted_blocking_rules.pop(0)]
        self.cover_set = self.cover_blocking_rules[0].training_coverage

        for blocking_rule in _sorted_blocking_rules:
            # check if the required recall is already reached:
            if len(self.cover_set) >= int(self.recall * self.full_set_size):
                break
            # check if subset is dominated by the cover_set:
            if blocking_rule.training_coverage.issubset(self.cover_set):
                continue
            self.cover_set = self.cover_set.union(blocking_rule.training_coverage)
            self.cover_blocking_rules.append(blocking_rule)

    def fit(self, sdf: DataFrame) -> 'BlockLearner':
        """
        This method fits, i.e. learns, the blocking rules that are needed to cover `recall` percent of the training
        set pairs. The fitting is done by solving the set-cover problem. It is solved by using a greedy algorithm.

        Args:
            sdf: a labelled training set containing pairs.

        Returns:
            the object itself
        """
        # verify whether `row_id` and `label` are columns of `sdf_1`
        if 'row_id' not in sdf.columns:
            raise AssertionError('`row_id` is not present as a column of sdf_1')

        if 'label' not in sdf.columns:
            raise AssertionError('`label` is not present as a column of sdf_1')


        # determine the full set of pairs in the training data that have positive labels from active learning:
        sdf = (
            sdf
            .filter(F.col('label') == 1)
            .persist()  # break the lineage to avoid recomputing since sdf_1 is used many times during the fitting
        )
        self.full_set = set(sdf.select('row_id').toPandas()['row_id'])
        # determine the cardinality of the full_set, i.e. |full_set|:
        self.full_set_size = len(self.full_set)

        # calculate the training coverage for each blocking rule:
        self.blocking_rules = (
            list(
                map(
                    lambda x: x.calculate_training_set_coverage(sdf),
                    self.blocking_rules
                )
            )
        )

        # use a greedy set cover algorithm to select a subset of the blocking rules that cover `recall` * |full_set|
        self._greedy_set_coverage()
        if self.verbose:
            print('Blocking rules:', ", ".join([x.__repr__() for x in self.cover_blocking_rules]))
        self.fitted = True
        return self

    def _create_blocks(self, sdf: DataFrame) -> List[DataFrame]:
        """
        This method creates a list of blocked data. Blocked data is created by applying the learned blocking rules on
        the input dataframe.

        Args:
            sdf: dataframe containing records

        Returns:
            A list of blocked dataframes, i.e. list of dataframes containing block-keys
        """
        sdf_blocks = []
        for blocking_rule in self.cover_blocking_rules:
            sdf_blocks.append(blocking_rule.create_block_key(sdf))
        return sdf_blocks

    @staticmethod
    def _create_block_table(blocks: List[DataFrame]) -> DataFrame:
        """
        This method unifies the blocked data into a single dataframe.
        Args:
            blocks: containing blocked dataframes, i.e. dataframes containing block-keys

        Returns:
            a unified dataframe with all the block-keys
        """
        block_table = blocks[0]
        for block in blocks[1:]:
            block_table = block_table.unionByName(block)
        return block_table

    def transform(self, sdf_1: DataFrame, sdf_2: Optional[DataFrame] = None) -> Union[DataFrame,
                                                                                      Tuple[DataFrame, DataFrame]]:
        """
        This method adds the block-keys to the input dataframes. It applies all the learned blocking rules on the
        input data and unifies the results. The result of this method is/are the input dataframe(s) containing the
        block-keys from the learned blocking rules.

        Args:
            sdf_1: dataframe containing records
            sdf_2: dataframe containing records

        Returns:
            dataframe(s) containing block-keys from the learned blocking-rules
        """
        if not self.fitted:
            raise ValueError('BlockLearner is not yet fitted')

        sdf_1_blocks = self._create_blocks(sdf_1)
        sdf_1_blocks = self._create_block_table(sdf_1_blocks)

        if sdf_2:
            sdf_1_blocks = self.table_checkpointer(sdf_1_blocks, checkpoint_name='sdf_1_blocks')
            sdf_2_blocks = self._create_blocks(sdf_2)
            sdf_2_blocks = self._create_block_table(sdf_2_blocks)

            return sdf_1_blocks, sdf_2_blocks

        return sdf_1_blocks



# Authors: Ahmet Bayraktar
#          Stan Leisink
#          Frits Hermans

from collections import defaultdict
from typing import List, Dict, Callable

import numpy as np
import pandas as pd
import scipy.spatial.distance as ssd
from scipy.cluster import hierarchy


def _perform_clustering(component: pd.DataFrame, threshold: float):
    """
    Apply hierarchical clustering to scored_pairs_table with component_ids
    Args:
        component: pandas dataframe containing all pairs and the similarity scores for a connected component
        threshold: threshold to apply in hierarchical clustering
    Returns:
        Generator that contains tuples of ids and scores
    """

    distance_threshold = 1 - threshold
    if len(component) > 1:
        i_to_id, condensed_distances = _get_condensed_distances(component)

        linkage = hierarchy.linkage(condensed_distances, method='centroid')
        partition = hierarchy.fcluster(linkage, t=distance_threshold, criterion='distance')

        clusters: Dict[int, List[int]] = defaultdict(list)

        for i, cluster_id in enumerate(partition):
            clusters[cluster_id].append(i)

        for cluster in clusters.values():
            if len(cluster) > 1:
                yield tuple(i_to_id[i] for i in cluster), None
    else:
        ids = np.array([int(component['row_number_1']), int(component['row_number_2'])])
        score = float(component['score'])
        if score > threshold:
            yield tuple(ids), (score,) * 2


def _convert_data_to_adjacency_matrix(component: pd.DataFrame):
    """
    This function converts a pd.DataFrame to a numpy adjacency matrix
    Args:
        component: pd.DataFrame
    Returns:
        index of elements of the components and a numpy adjacency matrix
    """
    def _get_adjacency_matrix(df, col1, col2, score_col:str = 'score'):
        df = pd.crosstab(df[col1], df[col2], values=df[score_col], aggfunc='max')
        idx = df.columns.union(df.index)
        df = df.reindex(index = idx, columns=idx, fill_value=0).fillna(0)
        return df

    a_to_b = _get_adjacency_matrix(component, "row_number_1", "row_number_2")
    b_to_a = _get_adjacency_matrix(component, "row_number_2", "row_number_1")

    symmetric_adjacency_matrix = a_to_b + b_to_a

    return symmetric_adjacency_matrix.index, np.array(symmetric_adjacency_matrix)


def _get_condensed_distances(component: pd.DataFrame):
    """
    Converts the pairwise list of distances to "condensed distance matrix" required by the hierarchical clustering
    algorithms. Also return a dictionary that maps the distance matrix to the ids.

    Args:
        component: pandas dataframe containing all pairs and the similarity scores for a connected component

    Returns:
        condensed distances and a dict with ids
    """

    i_to_id, adj_matrix = _convert_data_to_adjacency_matrix(component)
    distances = (np.ones_like(adj_matrix) - np.eye(len(adj_matrix))) - adj_matrix
    return dict(enumerate(i_to_id)), ssd.squareform(distances)


def _convert_dedupe_result_to_pandas_dataframe(dedupe_result: List, component_id: int) -> pd.DataFrame:
    """
    Function to convert the dedupe result into a pandas dataframe.
    E.g.
        dedupe_result = [((1, 2), array([0.96, 0.96])), ((3, 4, 5), array([0.95, 0.95, 0.95]))]

        returns

        | row_number | entity_identifier |
        | ---------- | ------------------- |
        | 1          | 1                   |
        | 2          | 1                   |
        | 3          | 2                   |
        | 4          | 2                   |
        | 5          | 2                   |

    Args:
        dedupe_result: the result with the deduplication results from the clustering
    Returns:
        pandas dataframe with row_number and entity_identifier
    """
    if len(dedupe_result) == 0:
        return pd.DataFrame(data={'row_number': [], 'entity_identifier': []})

    entity_identifier = 0
    df_list = []

    for ids, _ in dedupe_result:
        df_ = pd.DataFrame(data={'row_number': list(ids), 'entity_identifier': f"{component_id}_{entity_identifier}"})
        df_list.append(df_)
        entity_identifier += 1
    return pd.concat(df_list)


def apply_deduplication(cluster_score_threshold: float) -> Callable:
    """
    This function is a wrapper function to parameterize the _apply_deduplucation function with an extra parameter for
    the cluster_score_threshold. The cluster_score_threshold is a threshold that is used in the hierarchical clustering
    to determine which entities are linked.
    Args:
        cluster_score_threshold: a float in [0,1]
    Returns:
        a function, i.e. _apply_deduplication, that can be called as a Pandas udf
    """
    def _apply_deduplication(component: pd.DataFrame) -> pd.DataFrame:
        """
        This function applies deduplication on a component, i.e. a subgraph calculated by the connected components
        algorithm. This function is applied to a spark dataframe in a pandas udf to distribute the deduplication
        over a Spark cluster, component by component.
        Args:
            component: a pandas Dataframe
        Returns:
            a pandas Dataframe with the results from the hierarchical clustering of the deduplication
        """
        # check that component is a unique value, otherwise the deduplication goes wrong!
        assert len(component['component_id'].unique()) == 1
        component_id = component['component_id'][0]

        # perform the clustering:
        component = list(_perform_clustering(component, cluster_score_threshold))

        # convert the results to a dataframe:
        return _convert_dedupe_result_to_pandas_dataframe(component, component_id)

    return _apply_deduplication



# Authors: Ahmet Bayraktar
#          Stan Leisink
#          Frits Hermans

from typing import Optional, List, Dict

from pyspark.sql import DataFrame, functions as F
from pyspark.sql import SparkSession
from pyspark.sql import Window
from sklearn.exceptions import NotFittedError

from spark_matcher.blocker.blocking_rules import BlockingRule
from spark_matcher.matching_base.matching_base import MatchingBase
from spark_matcher.scorer.scorer import Scorer
from spark_matcher.table_checkpointer import TableCheckpointer


class Matcher(MatchingBase):
    """
    Matcher class to apply record linkage. Provide either the column names `col_names` using the default string
    similarity metrics or explicitly define the string similarity metrics in a dict `field_info` as in the example
    below. If `blocking_rules` is left empty, default blocking rules are used. Otherwise provide blocking rules as
    a list containing `BlockingRule` instances (see example below). The number of perfect matches used during
    training is set by `n_perfect_train_matches`.

    E.g.:

    from spark_matcher.blocker.blocking_rules import FirstNChars

    myMatcher = Matcher(spark_session, field_info={'name':[metric_function_1, metric_function_2],
                                                   'address:[metric_function_1, metric_function_3]},
                                                   blocking_rules=[FirstNChars('name', 3)])

    Args:
        spark_session: Spark session
        col_names: list of column names to use for matching
        field_info: dict of column names as keys and lists of string similarity metrics as values
        blocking_rules: list of `BlockingRule` instances
        n_train_samples: nr of pair samples to be created for training
        ratio_hashed_samples: ratio of hashed samples to be created for training, rest is sampled randomly
        n_perfect_train_matches: nr of perfect matches used for training
        scorer: a Scorer object used for scoring pairs
        verbose: sets verbosity
    """
    def __init__(self, spark_session: SparkSession, table_checkpointer: Optional[TableCheckpointer]=None,
                 checkpoint_dir: Optional[str]=None, col_names: Optional[List[str]] = None,
                 field_info: Optional[Dict] = None, blocking_rules: Optional[List[BlockingRule]] = None,
                 blocking_recall: float = 1.0, n_perfect_train_matches=1, n_train_samples: int = 100_000,
                 ratio_hashed_samples: float = 0.5, scorer: Optional[Scorer] = None, verbose: int = 0):
        super().__init__(spark_session, table_checkpointer, checkpoint_dir, col_names, field_info, blocking_rules,
                         blocking_recall, n_perfect_train_matches, n_train_samples, ratio_hashed_samples, scorer,
                         verbose)
        self.fitted_ = False

    def _create_predict_pairs_table(self, sdf_1_blocked: DataFrame, sdf_2_blocked: DataFrame) -> DataFrame:
        """
        Method to create pairs within blocks as provided by both input Spark dataframes in the column `block_key`.
        Note that an inner join is performed to create pairs: entries in `sdf_1_blocked` and `sdf_2_blocked`
        that don't share a `block_key` value in the other table will not appear in the resulting pairs table.

        Args:
            sdf_1_blocked: Spark dataframe containing all columns used for matching and the `block_key`
            sdf_2_blocked: Spark dataframe containing all columns used for matching and the `block_key`

        Returns:
            Spark dataframe containing all pairs to score

        """
        sdf_1_blocked = self._add_suffix_to_col_names(sdf_1_blocked, 1)
        sdf_2_blocked = self._add_suffix_to_col_names(sdf_2_blocked, 2)

        predict_pairs_table = (sdf_1_blocked.join(sdf_2_blocked, on='block_key', how='inner')
                               .drop_duplicates(subset=[col + "_1" for col in self.col_names] +
                                                       [col + "_2" for col in self.col_names])
                               )
        return predict_pairs_table

    def predict(self, sdf_1, sdf_2, threshold=0.5, top_n=None):
        """
        Method to predict on data used for training or new data.

        Args:
            sdf_1: first table
            sdf_2: second table
            threshold: probability threshold
            top_n: only return best `top_n` matches above threshold

        Returns:
            Spark dataframe with the matching result

        """
        if not self.fitted_:
            raise NotFittedError('The Matcher instance is not fitted yet. Call `fit` and train the instance.')

        sdf_1, sdf_2 = self._simplify_dataframe_for_matching(sdf_1), self._simplify_dataframe_for_matching(sdf_2)
        sdf_1_blocks, sdf_2_blocks = self.blocker.transform(sdf_1, sdf_2)
        pairs_table = self._create_predict_pairs_table(sdf_1_blocks, sdf_2_blocks)
        metrics_table = self._calculate_metrics(pairs_table)
        scores_table = (
            metrics_table
            .withColumn('score', self.scoring_learner.predict_proba(metrics_table['similarity_metrics']))
        )
        scores_table_filtered = (scores_table.filter(F.col('score') >= threshold)
                                 .drop('block_key', 'similarity_metrics'))

        if top_n:
            # we add additional columns to order by to remove ties and return exactly n items
            window = (Window.partitionBy(*[x + "_1" for x in self.col_names])
                      .orderBy(F.desc('score'), *[F.asc(col) for col in [x + "_2" for x in self.col_names]]))
            result = (scores_table_filtered.withColumn('rank', F.rank().over(window))
                      .filter(F.col('rank') <= top_n)
                      .drop('rank'))
        else:
            result = scores_table_filtered

        return result



# Authors: Ahmet Bayraktar
#          Stan Leisink
#          Frits Hermans

import abc
import os

from pyspark.sql import SparkSession, DataFrame


class TableCheckpointer(abc.ABC):
    """
    Args:
        spark_session: a spark session
        database: a name of a database or storage system where the tables can be saved
        checkpoint_prefix: a prefix of the name that can be used to save tables
    """

    def __init__(self, spark_session: SparkSession, database: str, checkpoint_prefix: str = "checkpoint_spark_matcher"):
        self.spark_session = spark_session
        self.database = database
        self.checkpoint_prefix = checkpoint_prefix

    def __call__(self, sdf: DataFrame, checkpoint_name: str):
        return self.checkpoint_table(sdf, checkpoint_name)

    @abc.abstractmethod
    def checkpoint_table(self, sdf: DataFrame, checkpoint_name: str):
        pass


class HiveCheckpointer(TableCheckpointer):
    """
    Args:
        spark_session: a spark session
        database: a name of a database or storage system where the tables can be saved
        checkpoint_prefix: a prefix of the name that can be used to save tables
    """
    def __init__(self, spark_session: SparkSession, database: str, checkpoint_prefix: str = "checkpoint_spark_matcher"):
        super().__init__(spark_session, database, checkpoint_prefix)

    def checkpoint_table(self, sdf: DataFrame, checkpoint_name: str):
        """
        This method saves the input dataframes as checkpoints of the algorithm. This checkpointing can be
        used to store intermediary results that are needed throughout the algorithm. The tables are stored using the
        following name convention: `{checkpoint_prefix}_{checkpoint_name}`.

        Args:
            sdf: a Spark dataframe that needs to be saved as a checkpoint
            checkpoint_name: name of the table

        Returns:
            the same, unchanged, spark dataframe as the input dataframe. With the only difference that the
            dataframe is now read from disk as a checkpoint.
        """
        sdf.write.saveAsTable(f"{self.database}.{self.checkpoint_prefix}_{checkpoint_name}",
                              mode="overwrite")
        sdf = self.spark_session.table(f"{self.database}.{self.checkpoint_prefix}_{checkpoint_name}")
        return sdf


class ParquetCheckPointer(TableCheckpointer):
    """
    Args:
        spark_session: a spark session
        checkpoint_dir: directory where the tables can be saved
        checkpoint_prefix: a prefix of the name that can be used to save tables
    """
    def __init__(self, spark_session: SparkSession, checkpoint_dir: str,
                 checkpoint_prefix: str = "checkpoint_spark_matcher"):
        super().__init__(spark_session, checkpoint_dir, checkpoint_prefix)

    def checkpoint_table(self, sdf: DataFrame, checkpoint_name: str):
        """
        This method saves the input dataframes as checkpoints of the algorithm. This checkpointing can be
        used to store intermediary results that are needed throughout the algorithm. The tables are stored
        using the
        following name convention: `{checkpoint_prefix}_{checkpoint_name}`.

        Args:
            sdf: a Spark dataframe that needs to be saved as a checkpoint
            checkpoint_name: name of the table

        Returns:
            the same, unchanged, spark dataframe as the input dataframe. With the only difference that the
            dataframe is now read from disk as a checkpoint.
        """
        file_name = os.path.join(f'{self.database}', f'{self.checkpoint_prefix}_{checkpoint_name}')
        sdf.write.parquet(file_name, mode='overwrite')
        return self.spark_session.read.parquet(file_name)



# Authors: Ahmet Bayraktar
#          Stan Leisink
#          Frits Hermans

from typing import List, Callable, Dict

import pandas as pd

from pyspark.sql import DataFrame, Column
from pyspark.sql import functions as F
from pyspark.sql import types as T


class SimilarityMetrics:
    """
    Class to calculate similarity metrics for pairs of records. The `field_info` dict contains column names as keys
    and lists of similarity functions as values. E.g.

    field_info = {'name': [token_set_ratio, token_sort_ratio],
                  'postcode': [ratio]}

    where `token_set_ratio`, `token_sort_ratio` and `ratio` are string similarity functions that take two strings
    as arguments and return a numeric value

    Attributes:
        field_info: dict containing column names as keys and lists of similarity functions as values
    """
    def __init__(self, field_info: Dict):
        self.field_info = field_info

    @staticmethod
    def _distance_measure_pandas(strings_1: List[str], strings_2: List[str], func: Callable) -> pd.Series:
        """
        Helper function to apply a string similarity metric to two arrays of strings. To be used in a Pandas UDF.

        Args:
            strings_1: array containing strings
            strings_2: array containing strings
            func: string similarity function to be applied

        Returns:
            Pandas series containing string similarities

        """
        df = pd.DataFrame({'x': strings_1, 'y': strings_2})
        return df.apply(lambda row: func(row['x'], row['y']), axis=1)

    @staticmethod
    def _create_similarity_metric_udf(similarity_metric_function: Callable):
        """
        Function that created Pandas UDF for a given similarity_metric_function

        Args:
            similarity_metric_function: function that takes two strings and returns a number

        Returns:
            Pandas UDF
        """

        @F.pandas_udf(T.FloatType())
        def similarity_udf(strings_1: pd.Series, strings_2: pd.Series) -> pd.Series:
            # some similarity metrics cannot deal with empty strings, therefore these are replaced with " "
            strings_1 = [x if x != "" else " " for x in strings_1]
            strings_2 = [x if x != "" else " " for x in strings_2]
            return SimilarityMetrics._distance_measure_pandas(strings_1, strings_2, similarity_metric_function)

        return similarity_udf

    def _apply_distance_metrics(self) -> Column:
        """
        Function to apply all distance metrics in the right order to string pairs and returns them in an array. This
        array is used as input to the (logistic regression) scoring function.

        Returns:
            array with the string distance metrics

        """
        distance_metrics_list = []
        for field_name in self.field_info.keys():
            field_name_1, field_name_2 = field_name + "_1", field_name + "_2"
            for similarity_function in self.field_info[field_name]:
                similarity_metric = (
                    SimilarityMetrics._create_similarity_metric_udf(similarity_function)(F.col(field_name_1),
                                                                                         F.col(field_name_2)))
                distance_metrics_list.append(similarity_metric)

        array = F.array(*distance_metrics_list)

        return array

    def transform(self, pairs_table: DataFrame) -> DataFrame:
        """
        Method to apply similarity metrics to pairs table. Method makes use of method dispatching to facilitate both
        Pandas and Spark dataframes

        Args:
            pairs_table: Spark or Pandas dataframe containing pairs table

        Returns:
            Pandas or Spark dataframe with pairs table and newly created `similarity_metrics` column

        """
        return pairs_table.withColumn('similarity_metrics', self._apply_distance_metrics())



# Authors: Ahmet Bayraktar
#          Stan Leisink
#          Frits Hermans

import abc

from pyspark.sql import Column, DataFrame, functions as F, types as T


class BlockingRule(abc.ABC):
    """
    Abstract class for blocking rules. This class contains all the base functionality for blocking rules.

    Attributes:
        blocking_column: the column on which the `BlockingRule` is applied
    """

    def __init__(self, blocking_column: str):
        self.blocking_column = blocking_column
        self.training_coverage = None
        self.training_coverage_size = None

    @abc.abstractmethod
    def _blocking_rule(self, c: Column) -> Column:
        """
        Abstract method for a blocking-rule

        Args:
            c: a Column

        Returns:
            A Column with a blocking-rule result
        """
        pass

    @abc.abstractmethod
    def __repr__(self) -> str:
        """
        Abstract method for a class representation

        Returns:
            return a string that represents the blocking-rule

        """
        pass

    def _apply_blocking_rule(self, c: Column) -> Column:
        """
        This method applies the blocking rule on an input column and adds a unique representation id to make sure the
        block-key is unique. Uniqueness is important to avoid collisions between block-keys, e.g. a blocking-rule that
        captures the first string character can return the same as a blocking-rule that captures the last character. To
        avoid this, the blocking-rule results is concatenated with the __repr__ method.

        Args:
            c: a Column

        Returns:
            a Column with a block_key

        """
        return F.concat(F.lit(f'{self.__repr__()}:'), self._blocking_rule(c))

    def create_block_key(self, sdf: DataFrame) -> DataFrame:
        """
        This method calculates and adds the block-key column to the input dataframe

        Args:
            sdf: a dataframe with records that need to be matched

        Returns:
            the dataframe with the block-key column
        """
        return sdf.withColumn('block_key', self._apply_blocking_rule(sdf[f'{self.blocking_column}']))

    def _create_training_block_keys(self, sdf: DataFrame) -> DataFrame:
        """
        This method is used to create block-keys on a training dataframe

        Args:
            sdf: a dataframe containing record pairs for training

        Returns:
            the dataframe containing block-keys for the record pairs
        """
        return (
            sdf
                .withColumn('block_key_1', self._apply_blocking_rule(sdf[f'{self.blocking_column}_1']))
                .withColumn('block_key_2', self._apply_blocking_rule(sdf[f'{self.blocking_column}_2']))
        )

    @staticmethod
    def _compare_and_filter_keys(sdf: DataFrame) -> DataFrame:
        """
        This method is used to compare block-keys of record pairs and subsequently filter record pairs in the training
        dataframe that have identical block-keys

        Args:
            sdf: a dataframe containing record pairs for training with their block-keys

        Returns:
            the filtered dataframe containing only record pairs with identical block-keys
        """
        return (
            sdf
                # split on `:` since this separates the `__repr__` from the blocking_rule result
                .withColumn('_blocking_result_1', F.split(F.col('block_key_1'), ":").getItem(1))
                .withColumn('_blocking_result_2', F.split(F.col('block_key_2'), ":").getItem(1))
                .filter(
                (F.col('block_key_1') == F.col('block_key_2')) &
                ((F.col('_blocking_result_1') != '') & (F.col('_blocking_result_2') != ''))
            )
                .drop('_blocking_result_1', '_blocking_result_2')
        )

    @staticmethod
    def _length_check(c: Column, n: int, word_count: bool = False) -> Column:
        """
        This method checks the length of the created block key.

        Args:
            c: block key to check
            n: given length of the string
            word_count: whether to check the string length or word count
        Returns:
            the block key if it is not shorter than the given length, otherwise returns None
        """
        if word_count:
            return F.when(F.size(c) >= n, c).otherwise(None)

        return F.when(F.length(c) == n, c).otherwise(None)

    def calculate_training_set_coverage(self, sdf: DataFrame) -> 'BlockingRule':
        """
        This method calculate the set coverage of the blocking rule on the training pairs. The set coverage of the rule
        is determined by looking at how many record pairs in the training set end up in the same block. This coverage
        is used in the BlockLearner to sort blocking rules in the greedy set_covering algorithm.
        Args:
            sdf: a dataframe containing record pairs for training

        Returns:
            The object itself
        """
        sdf = self._create_training_block_keys(sdf)

        sdf = self._compare_and_filter_keys(sdf)

        self.training_coverage = set(
            sdf
                .agg(F.collect_set('row_id').alias('training_coverage'))
                .collect()[0]['training_coverage']
        )

        self.training_coverage_size = len(self.training_coverage)
        return self


# define the concrete blocking rule examples:

class FirstNChars(BlockingRule):
    def __init__(self, blocking_column: str, n: int = 3):
        super().__init__(blocking_column)
        self.n = n

    def __repr__(self):
        return f"first_{self.n}_characters_{self.blocking_column}"

    def _blocking_rule(self, c: Column) -> Column:
        key = F.substring(c, 0, self.n)
        return self._length_check(key, self.n)


class FirstNCharsLastWord(BlockingRule):
    def __init__(self, blocking_column: str, n: int = 3, remove_non_alphanumerical=False):
        super().__init__(blocking_column)
        self.n = n
        self.remove_non_alphanumerical = remove_non_alphanumerical

    def __repr__(self):
        return f"first_{self.n}_characters_last_word_{self.blocking_column}"

    def _blocking_rule(self, c: Column) -> Column:
        if self.remove_non_alphanumerical:
            c = F.regexp_replace(c, r'\W+', ' ')
        tokens = F.split(c, r'\s+')
        last_word = F.element_at(tokens, -1)
        key = F.substring(last_word, 1, self.n)
        return self._length_check(key, self.n, word_count=False)


class FirstNCharactersFirstTokenSorted(BlockingRule):
    def __init__(self, blocking_column: str, n: int = 3, remove_non_alphanumerical=False):
        super().__init__(blocking_column)
        self.n = n
        self.remove_non_alphanumerical = remove_non_alphanumerical

    def __repr__(self):
        return f"first_{self.n}_characters_first_token_sorted_{self.blocking_column}"

    def _blocking_rule(self, c):
        if self.remove_non_alphanumerical:
            c = F.regexp_replace(c, r'\W+', ' ')
        tokens = F.split(c, r'\s+')
        sorted_tokens = F.sort_array(tokens)
        filtered_tokens = F.filter(sorted_tokens, lambda x: F.length(x) >= self.n)
        first_token = filtered_tokens.getItem(0)
        return F.substring(first_token, 1, self.n)


class LastNChars(BlockingRule):
    def __init__(self, blocking_column: str, n: int = 3):
        super().__init__(blocking_column)
        self.n = n

    def __repr__(self):
        return f"last_{self.n}_characters_{self.blocking_column}"

    def _blocking_rule(self, c: Column) -> Column:
        key = F.substring(c, -self.n, self.n)
        return self._length_check(key, self.n)


class WholeField(BlockingRule):
    def __init__(self, blocking_column: str):
        super().__init__(blocking_column)

    def __repr__(self):
        return f"whole_field_{self.blocking_column}"

    def _blocking_rule(self, c: Column) -> Column:
        return c


class FirstNWords(BlockingRule):
    def __init__(self, blocking_column: str, n: int = 1, remove_non_alphanumerical=False):
        super().__init__(blocking_column)
        self.n = n
        self.remove_non_alphanumerical = remove_non_alphanumerical

    def __repr__(self):
        return f"first_{self.n}_words_{self.blocking_column}"

    def _blocking_rule(self, c: Column) -> Column:
        if self.remove_non_alphanumerical:
            c = F.regexp_replace(c, r'\W+', ' ')
        tokens = F.split(c, r'\s+')
        key = self._length_check(tokens, self.n, word_count=True)
        return F.array_join(F.slice(key, 1, self.n), ' ')


class FirstNLettersNoSpace(BlockingRule):
    def __init__(self, blocking_column: str, n: int = 3):
        super().__init__(blocking_column)
        self.n = n

    def __repr__(self):
        return f"first_{self.n}_letters_{self.blocking_column}_no_space"

    def _blocking_rule(self, c: Column) -> Column:
        key = F.substring(F.regexp_replace(c, r'[^a-zA-Z]+', ''), 1, self.n)
        return self._length_check(key, self.n)


class SortedIntegers(BlockingRule):
    def __init__(self, blocking_column: str):
        super().__init__(blocking_column)

    def __repr__(self):
        return f"sorted_integers_{self.blocking_column}"

    def _blocking_rule(self, c: Column) -> Column:
        number_string = F.trim(F.regexp_replace(c, r'[^0-9\s]+', ''))
        number_string_array = F.when(number_string != '', F.split(number_string, r'\s+'))
        number_int_array = F.transform(number_string_array, lambda x: x.cast(T.IntegerType()))
        number_sorted = F.array_sort(number_int_array)
        return F.array_join(number_sorted, " ")


class FirstInteger(BlockingRule):
    def __init__(self, blocking_column: str):
        super().__init__(blocking_column)

    def __repr__(self):
        return f"first_integer_{self.blocking_column}"

    def _blocking_rule(self, c: Column) -> Column:
        number_string_array = F.split(F.trim(F.regexp_replace(c, r'[^0-9\s]+', '')), r'\s+')
        number_int_array = F.transform(number_string_array, lambda x: x.cast(T.IntegerType()))
        first_number = number_int_array.getItem(0)
        return first_number.cast(T.StringType())


class LastInteger(BlockingRule):
    def __init__(self, blocking_column: str):
        super().__init__(blocking_column)

    def __repr__(self):
        return f"last_integer_{self.blocking_column}"

    def _blocking_rule(self, c: Column) -> Column:
        number_string_array = F.split(F.trim(F.regexp_replace(c, r'[^0-9\s]+', '')), r'\s+')
        number_int_array = F.transform(number_string_array, lambda x: x.cast(T.IntegerType()))
        last_number = F.slice(number_int_array, -1, 1).getItem(0)
        return last_number.cast(T.StringType())


class LargestInteger(BlockingRule):
    def __init__(self, blocking_column: str):
        super().__init__(blocking_column)

    def __repr__(self):
        return f"largest_integer_{self.blocking_column}"

    def _blocking_rule(self, c: Column) -> Column:
        number_string_array = F.split(F.trim(F.regexp_replace(c, r'[^0-9\s]+', '')), r'\s+')
        number_int_array = F.transform(number_string_array, lambda x: x.cast(T.IntegerType()))
        largest_number = F.array_max(number_int_array)
        return largest_number.cast(T.StringType())


class NLetterAbbreviation(BlockingRule):
    def __init__(self, blocking_column: str, n: int = 3):
        super().__init__(blocking_column)
        self.n = n

    def __repr__(self):
        return f"{self.n}_letter_abbreviation_{self.blocking_column}"

    def _blocking_rule(self, c: Column) -> Column:
        words = F.split(F.trim(F.regexp_replace(c, r'[0-9]+', '')), r'\s+')
        first_letters = F.when(F.size(words) >= self.n, F.transform(words, lambda x: F.substring(x, 1, 1)))
        return F.array_join(first_letters, '')


# this is an example of a blocking rule that contains a udf with plain python code:

class UdfFirstNChar(BlockingRule):
    def __init__(self, blocking_column: str, n: int):
        super().__init__(blocking_column)
        self.n = n

    def __repr__(self):
        return f"udf_first_integer_{self.blocking_column}"

    def _blocking_rule(self, c: Column) -> Column:
        @F.udf
        def _rule(s: str) -> str:
            return s[:self.n]

        return _rule(c)


default_blocking_rules = [FirstNChars, FirstNCharsLastWord, LastNChars, WholeField, FirstNWords, FirstNLettersNoSpace,
                          SortedIntegers, FirstInteger, LastInteger, LargestInteger, NLetterAbbreviation]



# Authors: Ahmet Bayraktar
#          Stan Leisink
#          Frits Hermans

from typing import Union, Optional

import numpy as np
from pyspark.sql import SparkSession, types as T, functions as F, Column
from sklearn.base import BaseEstimator
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LogisticRegression


class Scorer:
    def __init__(self, spark_session: SparkSession, binary_clf: Optional[BaseEstimator] = None):
        self.spark_session = spark_session
        self.binary_clf = binary_clf
        if not self.binary_clf:
            self._create_default_clf()
        self.fitted_ = False

    def _create_default_clf(self) -> None:
        """
        This method creates a Sklearn Pipeline with a Standard Scaler and a Logistic Regression classifier.
        """
        self.binary_clf = (
            make_pipeline(
                StandardScaler(),
                LogisticRegression(class_weight='balanced')
            )
        )

    def fit(self, X: np.ndarray, y: np.ndarray) -> 'Scorer':
        """
        This method fits a clf model on input data `X` nd the binary targets `y`.

        Args:
            X: training data
            y: training targets, containing binary values

        Returns:
            The object itself
        """

        if len(set(y)) == 1:  # in case active learning only resulted in labels from one class
            return self

        self.binary_clf.fit(X, y)
        self.fitted_ = True
        return self

    def _predict_proba(self, X: np.ndarray) -> np.ndarray:
        """
        This method implements the code for predict_proba on a numpy array.
        This method is used to score all the pairs during training.
        """
        return self.binary_clf.predict_proba(X)

    def _predict_proba_spark(self, X: Column) -> Column:
        """
            This method implements the code for predict_proba on a spark column.
            This method is used to score all the pairs during inference time.
        """
        broadcasted_clf = self.spark_session.sparkContext.broadcast(self.binary_clf)

        @F.pandas_udf(T.FloatType())
        def _distributed_predict_proba(array):
            """
            This inner function defines the Pandas UDF for predict_proba on a spark cluster
            """
            return array.apply(lambda x: broadcasted_clf.value.predict_proba([x])[0][1])

        return _distributed_predict_proba(X)

    def predict_proba(self, X: Union[Column, np.ndarray]) -> Union[Column, np.ndarray]:
        """
        This method implements the abstract predict_proba method. It predicts the 'probabilities' of the target class
        for given input data `X`.

        Args:
            X: input data

        Returns:
            the predicted probabilities
        """
        if isinstance(X, Column):
            return self._predict_proba_spark(X)

        if isinstance(X, np.ndarray):
            return self._predict_proba(X)

        raise ValueError(f"{type(X)} is an unsupported datatype for X")



# Authors: Ahmet Bayraktar
#          Stan Leisink
#          Frits Hermans

import warnings
from typing import Optional, List, Dict

import dill
from pyspark.sql import DataFrame, functions as F, SparkSession
import numpy as np
import pandas as pd
from thefuzz.fuzz import token_set_ratio, token_sort_ratio

from spark_matcher.activelearner.active_learner import ScoringLearner
from spark_matcher.blocker.block_learner import BlockLearner
from spark_matcher.sampler.training_sampler import HashSampler, RandomSampler
from spark_matcher.scorer.scorer import Scorer
from spark_matcher.similarity_metrics import SimilarityMetrics
from spark_matcher.blocker.blocking_rules import BlockingRule, default_blocking_rules
from spark_matcher.table_checkpointer import TableCheckpointer, ParquetCheckPointer


class MatchingBase:

    def __init__(self, spark_session: SparkSession, table_checkpointer: Optional[TableCheckpointer] = None,
                 checkpoint_dir: Optional[str] = None, col_names: Optional[List[str]] = None,
                 field_info: Optional[Dict] = None, blocking_rules: Optional[List[BlockingRule]] = None,
                 blocking_recall: float = 1.0, n_perfect_train_matches=1, n_train_samples: int = 100_000,
                 ratio_hashed_samples: float = 0.5, scorer: Optional[Scorer] = None, verbose: int = 0):
        self.spark_session = spark_session
        self.table_checkpointer = table_checkpointer
        if not self.table_checkpointer and checkpoint_dir:
            self.table_checkpointer = ParquetCheckPointer(self.spark_session, checkpoint_dir, "checkpoint_deduplicator")
        else:
            warnings.warn(
                'Either `table_checkpointer` or `checkpoint_dir` should be provided. This instance can only be used '
                'when loading a previously saved instance.')

        if col_names:
            self.col_names = col_names
            self.field_info = {col_name: [token_set_ratio, token_sort_ratio] for col_name in
                               self.col_names}
        elif field_info:
            self.col_names = list(field_info.keys())
            self.field_info = field_info
        else:
            warnings.warn(
                'Either `col_names` or `field_info` should be provided. This instance can only be used when loading a '
                'previously saved instance.')
            self.col_names = ['']  # needed for instantiating ScoringLearner

        self.n_train_samples = n_train_samples
        self.ratio_hashed_samples = ratio_hashed_samples
        self.n_perfect_train_matches = n_perfect_train_matches
        self.verbose = verbose

        if not scorer:
            scorer = Scorer(self.spark_session)
        self.scoring_learner = ScoringLearner(self.col_names, scorer, verbose=self.verbose)

        self.blocking_rules = blocking_rules
        if not self.blocking_rules:
            self.blocking_rules = [blocking_rule(col) for col in self.col_names for blocking_rule in
                                   default_blocking_rules]
        self.blocker = BlockLearner(blocking_rules=self.blocking_rules, recall=blocking_recall,
                                    table_checkpointer=self.table_checkpointer, verbose=self.verbose)

    def save(self, path: str) -> None:
        """
        Save the current instance to a pickle file.

        Args:
            path: Path and file name of pickle file

        """
        to_save = self.__dict__

        # the spark sessions need to be removed as they cannot be saved and re-used later
        to_save['spark_session'] = None
        setattr(to_save['scoring_learner'].learner.estimator, 'spark_session', None)
        setattr(to_save['table_checkpointer'], 'spark_session', None)

        with open(path, 'wb') as f:
            dill.dump(to_save, f)

    def load(self, path: str) -> None:
        """
        Load a previously trained and saved Matcher instance.

        Args:
            path: Path and file name of pickle file

        """
        with open(path, 'rb') as f:
            loaded_obj = dill.load(f)

        # the spark session that was removed before saving needs to be filled with the spark session of this instance
        loaded_obj['spark_session'] = self.spark_session
        setattr(loaded_obj['scoring_learner'].learner.estimator, 'spark_session', self.spark_session)
        setattr(loaded_obj['table_checkpointer'], 'spark_session', self.spark_session)

        self.__dict__.update(loaded_obj)

    def _simplify_dataframe_for_matching(self, sdf: DataFrame) -> DataFrame:
        """
        Select only the columns used for matching and drop duplicates.
        """
        return sdf.select(*self.col_names).drop_duplicates()

    def _create_train_pairs_table(self, sdf_1: DataFrame, sdf_2: Optional[DataFrame] = None) -> DataFrame:
        """
        Create pairs that are used for training. Based on the given 'ratio_hashed_samples', part of the sample
        pairs are generated using MinHashLSH technique to create pairs that are more likely to be a match. Rest is
        sampled using random selection.

        Args:
            sdf_1: Spark dataframe containing the first table to with the input should be matched
            sdf_2: Optional: Spark dataframe containing the second table that should be matched to the first table

        Returns:
            Spark dataframe with sampled pairs that should be compared during training
        """
        n_hashed_samples = int(self.n_train_samples * self.ratio_hashed_samples)

        h_sampler = HashSampler(self.table_checkpointer, self.col_names, n_hashed_samples)
        hashed_pairs_table = self.table_checkpointer.checkpoint_table(h_sampler.create_pairs_table(sdf_1, sdf_2),
                                                                      checkpoint_name='minhash_pairs_table')

        # creating additional random samples to ensure that the exact number of n_train_samples
        # is obtained after dropping duplicated records
        n_random_samples = int(1.5 * (self.n_train_samples - hashed_pairs_table.count()))

        r_sampler = RandomSampler(self.table_checkpointer, self.col_names, n_random_samples)
        random_pairs_table = self.table_checkpointer.checkpoint_table(r_sampler.create_pairs_table(sdf_1, sdf_2),
                                                                      checkpoint_name='random_pairs_table')

        pairs_table = (
            random_pairs_table
            .unionByName(hashed_pairs_table)
            .withColumn('perfect_train_match', F.lit(False))
            .drop_duplicates()
            .limit(self.n_train_samples)
        )

        # add some perfect matches to assure there are two labels in the train data and the classifier can be trained
        perfect_matches = (pairs_table.withColumn('perfect_train_match', F.lit(True))
                           .limit(self.n_perfect_train_matches))
        for col in self.col_names:
            perfect_matches = perfect_matches.withColumn(col + "_1", F.col(col + "_2"))

        pairs_table = (perfect_matches.unionByName(pairs_table
                                                   .limit(self.n_train_samples - self.n_perfect_train_matches)))

        return pairs_table

    def _calculate_metrics(self, pairs_table: DataFrame) -> DataFrame:
        """
        Method to apply similarity metrics to pairs table.

        Args:
            pairs_table: Spark dataframe containing pairs table

        Returns:
            Spark dataframe with pairs table and newly created `similarity_metrics` column

        """
        similarity_metrics = SimilarityMetrics(self.field_info)
        return similarity_metrics.transform(pairs_table)

    def _create_blocklearning_input(self, metrics_table: pd.DataFrame, threshold: int = 0.5) -> DataFrame:
        """
        Method to collect data used for block learning. This data consists of the manually labelled pairs with label 1
        and the train pairs with a score above the `threshold`

        Args:
            metrics_table: Pandas dataframe containing the similarity metrics
            threshold: scoring threshold

        Returns:
            Spark dataframe with data for block learning

        """
        metrics_table['score'] = self.scoring_learner.predict_proba(np.array(metrics_table['similarity_metrics'].tolist()))[:, 1]
        metrics_table.loc[metrics_table['score'] > threshold, 'label'] = '1'

        # get labelled positives from activelearner
        positive_train_labels = (self.scoring_learner.train_samples[self.scoring_learner.train_samples['y'] == '1']
                                 .rename(columns={'y': 'label'}))
        metrics_table = pd.concat([metrics_table, positive_train_labels]).drop_duplicates(
            subset=[col + "_1" for col in self.col_names] + [col + "_2" for col in self.col_names], keep='last')

        metrics_table = metrics_table[metrics_table.label == '1']

        metrics_table['row_id'] = np.arange(len(metrics_table))
        return self.spark_session.createDataFrame(metrics_table)

    def _add_suffix_to_col_names(self, sdf: DataFrame, suffix: int):
        """
        This method adds a suffix to the columns that are used in the algorithm.
        This is done in order to do a join (with two dataframes with the same schema) to create the pairs table.
        """
        for col in self.col_names:
            sdf = sdf.withColumnRenamed(col, f"{col}_{suffix}")
        return sdf

    def fit(self, sdf_1: DataFrame, sdf_2: Optional[DataFrame] = None) -> 'MatchingBase':
        """
        Fit the MatchingBase instance on the two dataframes `sdf_1` and `sdf_2` using active learning. You will be
        prompted to enter whether the presented pairs are a match or not. Note that `sdf_2` is an optional argument.
        `sdf_2` is used for Matcher (i.e. matching one table to another). In the case of Deduplication, only providing
        `sdf_1` is sufficient, in that case `sdf_1` will be deduplicated.

        Args:
            sdf_1: Spark dataframe
            sdf_2: Optional: Spark dataframe

        Returns:
            Fitted MatchingBase instance

        """
        sdf_1 = self._simplify_dataframe_for_matching(sdf_1)
        if sdf_2:
            sdf_2 = self._simplify_dataframe_for_matching(sdf_2)

        pairs_table = self._create_train_pairs_table(sdf_1, sdf_2)
        metrics_table = self._calculate_metrics(pairs_table)
        metrics_table_pdf = metrics_table.toPandas()
        self.scoring_learner.fit(metrics_table_pdf)
        block_learning_input = self._create_blocklearning_input(metrics_table_pdf)
        self.blocker.fit(block_learning_input)
        self.fitted_ = True
        return self



# Authors: Ahmet Bayraktar
#          Stan Leisink
#          Frits Hermans

from typing import List, Optional, Union

import numpy as np
import pandas as pd
from modAL.models import ActiveLearner
from modAL.uncertainty import uncertainty_sampling
from pyspark.sql import DataFrame
from sklearn.base import BaseEstimator


class ScoringLearner:
    """
    Class to train a string matching model using active learning.
    Attributes:
        col_names: column names used for matching
        scorer: the scorer to be used in the active learning loop
        min_nr_samples: minimum number of responses required before classifier convergence is tested
        uncertainty_threshold: threshold on the uncertainty of the classifier during active learning,
            used for determining if the model has converged
        uncertainty_improvement_threshold: threshold on the uncertainty improvement of classifier during active
            learning, used for determining if the model has converged
        n_uncertainty_improvement: span of iterations to check for largest difference between uncertainties
        n_queries: maximum number of iterations to be done for the active learning session
        sampling_method: sampling method to be used for the active learning session
        verbose: sets verbosity
    """
    def __init__(self, col_names: List[str], scorer: BaseEstimator, min_nr_samples: int = 10,
                 uncertainty_threshold: float = 0.1, uncertainty_improvement_threshold: float = 0.01,
                 n_uncertainty_improvement: int = 5, n_queries: int = 9999, sampling_method=uncertainty_sampling,
                 verbose: int = 0):
        self.col_names = col_names
        self.learner = ActiveLearner(
            estimator=scorer,
            query_strategy=sampling_method
        )
        self.counter_total = 0
        self.counter_positive = 0
        self.counter_negative = 0
        self.min_nr_samples = min_nr_samples
        self.uncertainty_threshold = uncertainty_threshold
        self.uncertainty_improvement_threshold = uncertainty_improvement_threshold
        self.n_uncertainty_improvement = n_uncertainty_improvement
        self.uncertainties = []
        self.n_queries = n_queries
        self.verbose = verbose

    def _input_assert(self, message: str, choices: List[str]) -> str:
        """
        Adds functionality to the python function `input` to limit the choices that can be returned
        Args:
            message: message to user
            choices: list containing possible choices that can be returned
        Returns:
            input returned by user
        """
        output = input(message).lower()
        if output not in choices:
            print(f"Wrong input! Your input should be one of the following: {', '.join(choices)}")
            return self._input_assert(message, choices)
        else:
            return output

    def _get_uncertainty_improvement(self) -> Optional[float]:
        """
        Calculates the uncertainty differences during active learning. The largest difference over the `last_n`
        iterations is returned. The aim of this function is to suggest early stopping of active learning.

        Returns: largest uncertainty update in `last_n` iterations

        """
        uncertainties = np.asarray(self.uncertainties)
        abs_differences = abs(uncertainties[1:] - uncertainties[:-1])
        return max(abs_differences[-self.n_uncertainty_improvement:])

    def _is_converged(self) -> bool:
        """
        Checks whether the model is converged by comparing the last uncertainty value with the `uncertainty_threshold`
        and comparing the `last_n` uncertainty improvements with the `uncertainty_improvement_threshold`. These checks
        are only performed if at least `min_nr_samples` are labelled.

        Returns:
            boolean indicating whether the model is converged

        """
        if (self.counter_total >= self.min_nr_samples) and (
                len(self.uncertainties) >= self.n_uncertainty_improvement + 1):
            uncertainty_improvement = self._get_uncertainty_improvement()
            if (self.uncertainties[-1] <= self.uncertainty_threshold) or (
                    uncertainty_improvement <= self.uncertainty_improvement_threshold):
                return True
        else:
            return False

    def _get_active_learning_input(self, query_inst: pd.DataFrame) -> np.ndarray:
        """
        Obtain user input for a query during active learning.
        Args:
            query_inst: query as provided by the ActiveLearner instance
        Returns: label of user input '1' or '0' as yes or no
                    'p' to go to previous
                    'f' to finish
                    's' to skip the query
        """
        print(f'\nNr. {self.counter_total + 1} ({self.counter_positive}+/{self.counter_negative}-)')
        print("Is this a match? (y)es, (n)o, (p)revious, (s)kip, (f)inish")
        print('')
        for element in [1, 2]:
            for col_name in self.col_names:
                print(f'{col_name}_{element}' + ': ' + query_inst[f'{col_name}_{element}'].iloc[0])
            print('')
        user_input = self._input_assert("", ['y', 'n', 'p', 'f', 's'])
        # replace 'y' and 'n' with '1' and '0' to make them valid y labels
        user_input = user_input.replace('y', '1').replace('n', '0')

        y_new = np.array([user_input])
        return y_new

    def _calculate_uncertainty(self, x) -> None:
        # take the maximum probability of the predicted classes as proxy of the confidence of the classifier
        confidence = self.predict_proba(x).max(axis=1)[0]
        if self.verbose:
            print('uncertainty:', 1 - confidence)
        self.uncertainties.append(1 - confidence)

    def _show_min_max_scores(self, X: pd.DataFrame) -> None:
        """
        Prints the lowest and the highest logistic regression scores on train data during active learning.

        Args:
            X: Pandas dataframe containing train data that is available for labelling duringg active learning
        """
        X_all = pd.concat((X, self.train_samples))
        pred_max = self.learner.predict_proba(np.array(X_all['similarity_metrics'].tolist())).max(axis=0)
        print(f'lowest score: {1 - pred_max[0]:.3f}')
        print(f'highest score: {pred_max[1]:.3f}')

    def _label_perfect_train_matches(self, identical_records: pd.DataFrame) -> None:
        """
        To prevent asking labels for the perfect matches that were created by setting `n_perfect_train_matches`, these
        are provided to the active learner upfront.

        Args:
            identical_records: Pandas dataframe containing perfect matches

        """
        identical_records['y'] = '1'
        self.learner.teach(np.array(identical_records['similarity_metrics'].values.tolist()),
                           identical_records['y'].values)
        self.train_samples = pd.concat([self.train_samples, identical_records])

    def fit(self, X: pd.DataFrame) -> 'ScoringLearner':
        """
        Fit ScoringLearner instance on pairs of strings
        Args:
            X: Pandas dataframe containing pairs of strings and distance metrics of paired strings
        """
        self.train_samples = pd.DataFrame([])
        query_inst_prev = None

        # automatically label all perfect train matches:
        identical_records = X[X['perfect_train_match']].copy()
        self._label_perfect_train_matches(identical_records)
        X = X.drop(identical_records.index).reset_index(drop=True)  # remove identical records to avoid double labelling

        for i in range(self.n_queries):
            query_idx, query_inst = self.learner.query(np.array(X['similarity_metrics'].tolist()))

            if self.learner.estimator.fitted_:
                # the uncertainty calculations need a fitted estimator
                # however it can occur that the estimator can only be fit after a couple rounds of querying
                self._calculate_uncertainty(query_inst)
                if self.verbose >= 2:
                    self._show_min_max_scores(X)

            y_new = self._get_active_learning_input(X.iloc[query_idx])
            if y_new == 'p':  # use previous (input is 'p')
                y_new = self._get_active_learning_input(query_inst_prev)
            elif y_new == 'f':  # finish labelling (input is 'f')
                break
            query_inst_prev = X.iloc[query_idx]
            if y_new != 's':  # skip case (input is 's')
                self.learner.teach(np.asarray([X.iloc[query_idx]['similarity_metrics'].iloc[0]]), np.asarray(y_new))
                train_sample_to_add = X.iloc[query_idx].copy()
                train_sample_to_add['y'] = y_new
                self.train_samples = pd.concat([self.train_samples, train_sample_to_add])

            X = X.drop(query_idx).reset_index(drop=True)

            if self._is_converged():
                print("Classifier converged, enter 'f' to stop training")

            if y_new == '1':
                self.counter_positive += 1
            elif y_new == '0':
                self.counter_negative += 1
            self.counter_total += 1
        return self

    def predict_proba(self, X: Union[DataFrame, pd.DataFrame]) -> Union[DataFrame, pd.DataFrame]:
        """
        Predict probabilities on new data whether the pairs are a match or not
        Args:
            X: Pandas or Spark dataframe to predict on
        Returns: match probabilities
        """
        return self.learner.estimator.predict_proba(X)



from setuptools import setup, find_packages

base_packages = [
    'pandas',
    'numpy',
    'scikit-learn',
    'python-Levenshtein',
    'thefuzz',
    'modAL',
    'pytest',
    'multipledispatch',
    'dill',
    'graphframes',
    'scipy'
]

doc_packages = [
    'sphinx',
    'nbsphinx',
    'sphinx_rtd_theme'
]

util_packages = [
    'pyspark',
    'pyarrow',
    'jupyterlab'
]

base_doc_packages = base_packages + doc_packages
dev_packages = base_packages + doc_packages + util_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(name='Spark-Matcher',
      version='0.2',
      author="Ahmet Bayraktar, Stan Leisink, Frits Hermans",
      description="Record matching and entity resolution at scale in Spark",
      long_description=long_description,
      long_description_content_type="text/markdown",
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
      ],
      packages=find_packages(exclude=['examples']),
      package_data={"spark_matcher": ["data/*.csv"]},
      install_requires=base_packages,
      extras_require={
          "base": base_packages,
          "doc": base_doc_packages,
          "dev": dev_packages,
      },
      python_requires=">=3.7",
      )



# Authors: Ahmet Bayraktar
#          Stan Leisink
#          Frits Hermans

from typing import List

import numpy as np
import pandas as pd
from pyspark.ml.feature import StopWordsRemover
from pyspark.sql import DataFrame
from pyspark.sql import functions as F


def get_most_frequent_words(sdf: DataFrame, col_name: str, min_df=2, top_n_words=1_000) -> pd.DataFrame:
    """
    Count word frequencies in a Spark dataframe `sdf` column named `col_name` and return a Pandas dataframe containing
    the document frequencies of the `top_n_words`. This function is intended to be used to create a list of stopwords.

    Args:
        sdf: Spark dataframe
        col_name: column name to get most frequent words from
        min_df: minimal document frequency for a word to be included as stopword, int for number and float for fraction
        top_n_words: number of most occurring words to include

    Returns:
        pandas dataframe containing most occurring words, their counts and document frequencies

    """
    sdf_col_splitted = sdf.withColumn(f'{col_name}_array', F.split(F.col(col_name), pattern=' '))
    word_count = sdf_col_splitted.select(F.explode(f'{col_name}_array').alias('words')).groupBy('words').count()
    doc_count = sdf.count()
    word_count = word_count.withColumn('df', F.col('count') / doc_count)
    if isinstance(min_df, int):
        min_count = min_df
    elif isinstance(min_df, float):
        min_count = np.ceil(min_df * doc_count)
    word_count_pdf = word_count.filter(F.col('count') >= min_count).sort(F.desc('df')).limit(top_n_words).toPandas()
    return word_count_pdf


def remove_stopwords(sdf: DataFrame, col_name: str, stopwords: List[str], case=False,
                     suffix: str = '_wo_stopwords') -> DataFrame:
    """
    Remove stopwords `stopwords` from a column `col_name` in a Spark dataframe `sdf`. The result will be written to a
    new column, named with the concatenation of `col_name` and `suffix`.

    Args:
        sdf: Spark dataframe
        col_name: column name to remove stopwords from
        stopwords: list of stopwords to remove
        case: whether to check for stopwords including lower- or uppercase
        suffix: suffix for the newly created column

    Returns:
        Spark dataframe with column added without stopwords

    """
    sdf = sdf.withColumn(f'{col_name}_array', F.split(F.col(col_name), pattern=' '))
    sw_remover = StopWordsRemover(inputCol=f'{col_name}_array', outputCol=f'{col_name}_array_wo_stopwords',
                                  stopWords=stopwords, caseSensitive=case)
    sdf = (sw_remover.transform(sdf)
           .withColumn(f'{col_name}{suffix}', F.concat_ws(' ', F.col(f'{col_name}_array_wo_stopwords')))
           .fillna({f'{col_name}{suffix}': ''})
           .drop(f'{col_name}_array', f'{col_name}_array_wo_stopwords')
           )
    return sdf



# Authors: Ahmet Bayraktar
#          Stan Leisink
#          Frits Hermans

from typing import List, Tuple

from graphframes import GraphFrame
from pyspark.sql import functions as F, types as T, DataFrame, Window

from spark_matcher.table_checkpointer import TableCheckpointer


class ConnectedComponentsCalculator:

    def __init__(self, scored_pairs_table: DataFrame, max_edges_clustering: int,
                 edge_filter_thresholds: List[float], table_checkpointer: TableCheckpointer):
        self.scored_pairs_table = scored_pairs_table
        self.max_edges_clustering = max_edges_clustering
        self.edge_filter_thresholds = edge_filter_thresholds
        self.table_checkpointer = table_checkpointer

    @staticmethod
    def _create_graph(scores_table: DataFrame) -> GraphFrame:
        """
        This function creates a graph where each row-number is a vertex and each similarity score between a pair of
        row-numbers represents an edge. This Graph is used as input for connected components calculations to determine
        the subgraphs of linked pairs.

        Args:
            scores_table: a pairs table with similarity scores for each pair
        Returns:
            a GraphFrames graph object representing the row-numbers and the similarity scores between them as a graph
        """

        vertices_1 = scores_table.select('row_number_1').withColumnRenamed('row_number_1', 'id')
        vertices_2 = scores_table.select('row_number_2').withColumnRenamed('row_number_2', 'id')
        vertices = vertices_1.unionByName(vertices_2).drop_duplicates()

        edges = (
            scores_table
                .select('row_number_1', 'row_number_2', 'score')
                .withColumnRenamed('row_number_1', 'src')
                .withColumnRenamed('row_number_2', 'dst')
        )
        return GraphFrame(vertices, edges)

    def _calculate_connected_components(self, scores_table: DataFrame, checkpoint_name: str) -> DataFrame:
        """
        This function calculates the connected components (i.e. the subgraphs) of a graph of scored pairs.
        The result of the connected-components algorithm is cached and saved.

        Args:
            scores_table: a pairs table with similarity scores for each pair
        Returns:
            a spark dataframe containing the connected components of a graph of scored pairs.
        """
        graph = self._create_graph(scores_table)
        connected_components = graph.connectedComponents()
        return self.table_checkpointer(connected_components, checkpoint_name)

    @staticmethod
    def _add_component_id_to_scores_table(scores_table: DataFrame, connected_components: DataFrame) -> DataFrame:
        """
        This function joins the initial connected-component identifiers to the scored pairs table. For each scored pair,
        a component identifier indicating to which subgraph a pair belongs is added.

        Args:
            scores_table: a pairs table with similarity scores for each pair
            connected_components: a spark dataframe containing the result of the connected components algorithm
        Returns:
            the scores_table with an identifier that indicates to which component a pair belongs
        """
        scores_table_with_component_ids = (
            scores_table
            .join(connected_components, on=scores_table['row_number_1']==connected_components['id'], how='left')
            .withColumnRenamed('component', 'component_1')
            .drop('id')
            .join(connected_components, on=scores_table['row_number_2']==connected_components['id'], how='left')
            .withColumnRenamed('component', 'component_2')
            .drop('id')
        )

        # verify the subgraphs of the connected components algorithm
        assert scores_table_with_component_ids.filter(F.col('component_1') != F.col('component_2')).count() == 0

        scores_table_with_component_ids = (
            scores_table_with_component_ids
            .withColumnRenamed('component_1', 'component')
            .drop('component_2')
        )
        return scores_table_with_component_ids

    @staticmethod
    def _add_big_components_info(scored_pairs_w_components: DataFrame, component_cols: List[str], max_size: int) -> \
    Tuple[DataFrame, bool]:
        """
        This function adds information if there are components in the scored_pairs table that contain more pairs than
        the `max_size'. If there is a component that is too big, the function will identify this and will indicate that
        filter iterations are required via the `continue_iteration` value.

        Args:
            scored_pairs_w_components: a pairs table with similarity scores for each pair and component identifiers
            component_cols: a list of columns that represent the composite key that indicate connected-components
            after one or multiple iterations
            max_size: a number indicating the maximum size of the number of pairs that are in a connected component
            in the scores table
        Returns:
            the scores table with a column indicating whether a pair belongs to a big component
            and a boolean indicating whether there are too big components and thus whether more iterations are required.
        """
        window = Window.partitionBy(*component_cols)
        big_components_info = (
            scored_pairs_w_components
                .withColumn('constant', F.lit(1))
                .withColumn('component_size', F.count('constant').over(window))
                .drop('constant')
                .withColumn('is_in_big_component', F.col('component_size') > max_size)
                .drop('component_size')
        )
        continue_iteration = False
        if big_components_info.filter(F.col('is_in_big_component')).count() > 0:
            continue_iteration = True
        return big_components_info, continue_iteration

    @staticmethod
    def _add_low_score_info(scored_pairs: DataFrame, threshold) -> DataFrame:
        """
        This function adds a column that indicates whether a similarity score between pairs is lower than a `threshold`
        value.

        Args:
            scored_pairs: a pairs table with similarity scores for each pair
            threshold: a value indicating the treshold value for a similarity score
        Returns:
            the scores table with an addiction column indicating whether a pair has a score lower than the `threshold`
        """
        scored_pairs = (
            scored_pairs
                .withColumn('is_lower_than_threshold', F.col('score') <threshold)
        )
        return scored_pairs

    @staticmethod
    def _join_connected_components_iteration_results(scored_pairs: DataFrame, connected_components: DataFrame,
                                                     iteration_number: int) -> DataFrame:
        """
        This function joins the the connected component results for each iteration to the scores table and prunes the
        scores table by removing edges that belong to a too big component and have a similarity score that is lower than
        the current threshold.

        Args:
            scored_pairs: a pairs table with similarity scores for each pair for an iteration
            connected_components: a spark dataframe containing the result of the connected components algorithm for
            an iteration
            iteration_number: a counter indicating which iteration it is
        Returns:
            a scores table with the component identifiers for an iteration and without edges that belonged to too big
            components
            that had scores that were below the current threshold.
        """
        # join the connected components results to the scored-pairs on `row_number_1`
        result = (
            scored_pairs
                .join(connected_components, on=scored_pairs['row_number_1'] == connected_components['id'], how='left')
                .drop('id')
                .withColumnRenamed('component', 'component_1')
                .persist())

        # join the connected components results to the scored-pairs on `row_number_2`
        result = (
            result
                .join(connected_components, on=scored_pairs['row_number_2'] == connected_components['id'], how='left')
                .drop('id')
                .withColumnRenamed('component', 'component_2')
                .filter((F.col('component_1') == F.col('component_2')) | (F.col('component_1').isNull() & F.col(
                'component_2').isNull()))
                .withColumnRenamed('component_1', f'component_iteration_{iteration_number}')
                .drop('component_2')
                # add a default value for rows that were not part of the iteration, i.e. -1 since component ids' are
                # always positive integers
                .fillna(-1, subset=[f'component_iteration_{iteration_number}'])
                .filter((F.col('is_in_big_component') & ~F.col('is_lower_than_threshold')) |
                        (~F.col('is_in_big_component')))
                .drop('is_in_big_component', 'is_lower_than_threshold')
        )
        return result

    @staticmethod
    def _add_component_id(sdf: DataFrame, iteration_count: int) -> DataFrame:
        """
        This function adds a final `component_id` after all iterations that can be used as a groupby-key to
        distribute the deduplication. The `component_id` is a hashed value of the concatenation with '_' seperation
        of all component identifiers of the iterations. The sha2 hash function is used order to make the likelihood of
        hash collisions negligibly small.

        Args:
            sdf: a dataframe containing the scored pairs
            iteration_count: a number indicating the final iteration number that is used.
        Returns:
            the final scores-table with a component-id that can be used as groupby-key to distribute the deduplication.
        """
        if iteration_count > 0:
            component_id_cols = ['initial_component'] + [f'component_iteration_{i}' for i in
                                                         range(1, iteration_count + 1)]
            sdf = sdf.withColumn('component_id', F.sha2(F.concat_ws('_', *component_id_cols), numBits=256))
        else:
            # use the initial component as component_id and cast the value to string to match with the return type of
            # the sha2 algorithm
            sdf = (
                sdf
                .withColumn('component_id', F.col('initial_component').cast(T.StringType()))
                .drop('initial_component')
                )
        return sdf

    def _create_component_ids(self) -> DataFrame:
        """
        This function wraps the other methods in this class and performs the connected-component calculations and the
        edge filtering if required.

        Returns:
            a scored-pairs table with a component identifier that can be used as groupby-key to distribute the
            deduplication.
        """
        # calculate the initial connected components
        connected_components = self._calculate_connected_components(self.scored_pairs_table,
                                                                    "cached_connected_components_table_initial")
        # add the component ids to the scored-pairs table
        self.scored_pairs_table = self._add_component_id_to_scores_table(self.scored_pairs_table, connected_components)
        self.scored_pairs_table = self.scored_pairs_table.withColumnRenamed('component', 'initial_component')

        #checkpoint the results
        self.scored_pairs_table = self.table_checkpointer(self.scored_pairs_table,
                                                          "cached_scores_with_components_table_initial")

        component_columns = ['initial_component']
        iteration_count = 0
        for i in range(1, len(self.edge_filter_thresholds) + 1):
            # set and update the threshold
            threshold = self.edge_filter_thresholds[i - 1]

            self.scored_pairs_table, continue_iter = self._add_big_components_info(self.scored_pairs_table,
                                                                                   component_columns,
                                                                                   self.max_edges_clustering)
            if not continue_iter:
                # no further iterations are required
                break

            self.scored_pairs_table = self._add_low_score_info(self.scored_pairs_table, threshold)
            # recalculate the connected_components
            iteration_input_sdf = (
                self.scored_pairs_table
                .filter((F.col('is_in_big_component') & ~F.col('is_lower_than_threshold')))
            )

            print(f"calculate connected components iteration {i} with edge filter threshold {threshold}")
            connected_components_iteration = self._calculate_connected_components(iteration_input_sdf,
                                                                f'cached_connected_components_table_iteration_{i}')

            self.scored_pairs_table = (
                self
                    ._join_connected_components_iteration_results(self.scored_pairs_table,
                                                                  connected_components_iteration, i)
            )

            # checkpoint the result to break the lineage after each iteration and to inspect intermediate results
            self.scored_pairs_table = self.table_checkpointer(self.scored_pairs_table,
                                                                f'cached_scores_with_components_table_iteration_{i}')

            component_columns.append(f'component_iteration_{i}')
            iteration_count += 1

        # final check to see if there were sufficient iterations to reduce the size of the big components if all
        # specified iterations are used to reduce the component size
        if iteration_count == len(self.edge_filter_thresholds):
            self.scored_pairs_table, continue_iter = self._add_big_components_info(self.scored_pairs_table,
                                                                                   component_columns,
                                                                                   self.max_edges_clustering)
            if continue_iter:
                print("THE EDGE-FILTER-THRESHOLDS ARE NOT ENOUGH TO SUFFICIENTLY REDUCE THE SIZE OF THE BIG COMPONENTS")

        # add the final component-id that can be used as groupby key for distributed deduplication
        self.scored_pairs_table = self._add_component_id(self.scored_pairs_table, iteration_count)

        return self.table_checkpointer(self.scored_pairs_table, 'cached_scores_with_components_table')
